/*****************************************************************************
* GROM Version 0.6                                                           *
* Bart Trzynadlowski, 2000, public domain.                                   *
*****************************************************************************/

/*
 * grom.c: 
 *
 * Simple utility for toying with Genesis ROMs. Supports converting of SMD to
 * BIN ROM formats. Also displays information from the Sega cartridge header.
 *
 * Much thanks to Kuwanger (http://members.tripod.com/~Kuwanger) for figuring
 * out that the documentation I was using about the SMD format was incorrect.
 * Odd bytes come BEFORE even bytes. Go visit his page, damnit. Also, XnaK for
 * writing his Genesis ROM text, it provided invaluable information on the
 * Sega cartridge header.
 */

#include <stdio.h>
#include <string.h>

#define VERSION "0.6"
#define error(str)                              \
        {                                       \
                printf(str);                    \
                return 0;                       \
        }

/* functions */
int convert_smd_bin(char *);
int show_info(char *);
void smd_bin(unsigned char *, unsigned char *);

/*****************************************************************************
 int main(int, char **):
                Stop reading this source code if you don't know what main()
                is. Your head might pop.
 Input:         Obvious.
 Returns:       Obvious.
*****************************************************************************/
int main(int argc, char **argv)
{
        int i, j;

        /* show help */
        if (argc <= 1)
        {
                printf("GROM Version %s by Bart Trzynadlowski, 2000\n",
                       VERSION);
                printf("A Free Multi-function Sega Genesis ROM Utility\n");
                printf("\n");
                printf("Usage:   grom [files] [options]\n");
                printf("File:    Genesis ROM file, BIN or SMD format\n");
                printf("Options: -bin\tConvert file(s) to BIN format\n");
                printf("         -info\tShow information about file(s)\n");
                return 0;
        }
        else if (argc <= 2)     /* user didn't specify an option */
        {
                printf("main: Not enough arguments\n");
                return 0;
        }

        /* actually process arguments, as each one is processed, we remove it
           from the argv[] array. */
        for (i = 1; i < argc; i++)
        {
                if (strcmp(argv[i], "-bin") == 0)       /* SMD to BIN */
                {
                        /* convert each file to bin */
                        for (j = 1; j < argc; j++)
                        {
                                if (*argv[j] == '-')    /* - means option */
                                        ;
                                /* if already BIN, let user know */
                                else if (strstr(argv[j], ".bin") != NULL)
                                        printf("main: Already in BIN format: %s\n", argv[j]);
                                /* if SMD, print message... */
                                else if (strstr(argv[j], ".smd") != NULL)
                                        convert_smd_bin(argv[j]);
                                else
                                        printf("main: Unrecognized format: %s\n", argv[j]);
                        }
                }
                else if (strcmp(argv[i], "-info") == 0) /* show info */
                {
                        /* print information on each file */
                        for (j = 1; j < argc; j++)
                        {
                                if (*argv[j] == '-')
                                        ;
                                else
                                        show_info(argv[j]);
                        }
                }
        }                

        return 0;
}

/*****************************************************************************
 int convert_smd_bin(char *):
                Converts an SMD ROM file (<filename>.smd) to a BIN ROM file
                (<filename>.bin).
 Input:         File path.
 Returns:       Nothing.
*****************************************************************************/
int convert_smd_bin(char *file)
{
        int i, num_blocks, meter = 0;
        FILE *in_fp, *out_fp;
        unsigned char smd_block[16384];
        unsigned char bin_block[16384];

        /* open SMD file, open BIN file, get number of blocks, read header */
        if ((in_fp = fopen(file, "rb")) == NULL)
                error("convert_smd_bin: SMD file could not be opened\n");
        if ((out_fp = fopen(strcat(strtok(file, "."), ".bin"), "wb")) == NULL)
                error("convert_smd_bin: BIN file could not be opened\n");                      
        fseek(in_fp, 0, SEEK_END);
        fgetpos(in_fp, (void *) &num_blocks);
        num_blocks = ((num_blocks + 1) - 512) / 16384;
        fseek(in_fp, 512, SEEK_SET);
       
        /* turn off line buffering on stdout so we can see progress meter */
        setvbuf(stdout, NULL, _IONBF, NULL);

        printf("convert_smd_bin: Converting %s to BIN format", strcat(strtok(file, "."), ".smd"));

        /* convert smd to bin */
        for (i = 0; i < num_blocks; i++)
        {
                fread(smd_block, 1, 16384, in_fp);
                smd_bin(bin_block, smd_block);
                fwrite(bin_block, 1, 16384, out_fp);

                /* increase meter every 1/10th total blocks */
                meter++;
                if (meter == (num_blocks / 10))
                {
                        printf(".");
                        meter = 0;
                }
        }

        printf(" done\n");

        setvbuf(stdout, NULL, _IOLBF, NULL);    /* line buffering back on */

        /* close files and exit */
        fclose(in_fp);
        fclose(out_fp);

        return 0;
}

/*****************************************************************************
 int show_info(char *file):
                Shows information about a given file. If it is a binary file
                information will be obtained from the Sega header, otherwise
                we will show whatever information we can obtain.
 Input:         File path.
 Returns:       Nothing.
*****************************************************************************/
int show_info(char *file)
{
        FILE *fp;
        unsigned char header[16384];
        unsigned char smd[16384];
        char string[49];

        /* check to see if it a BIN or SMD file */
        if ((strstr(file, ".bin") != NULL) || (strstr(file, ".smd") != NULL))
        {
                /* open file and read the header */
                if ((fp = fopen(file, "rb")) == NULL)
                        error("show_info: File could not be opened\n");

                if (strstr(file, ".bin") != NULL)       /* bin file */
                        fread(header, 1, 512, fp);
                else                                    /* smd file */
                {
                        /* get first block of SMD */
                        fread(smd, 1, 512, fp);         /* skip header */
                        fread(smd, 1, 16384, fp);
                        smd_bin(header, smd);           /* decode */
                }

                printf("Showing information from cartridge data...\n");

                /* print info from Sega header */
                printf("                    File: %s\n", file);
                string[16] = '\0';
                strncpy(string, &header[0x100], 16);
                printf("                  System: %s\n", string);
                strncpy(string, &header[0x110], 16);
                printf("               Copyright: %s\n", string);
                string[48] = '\0';
                strncpy(string, &header[0x120], 48);
                printf("    Game name (domestic): %s\n", string);
                strncpy(string, &header[0x150], 48);
                printf("    Game name (overseas): %s\n", string);
                if (header[0x180] == 'G' && header[0x181] == 'M')
                        printf("           Software type: game\n");
                else if (header[0x180] == 'A' && header[0x181] == 'l')
                        printf("           Software type: educational\n");
                else
                        printf("           Software type: %c%c",
                               header[0x180], header[0x181]);
                /* from personal observation, it seems the product code
                   field starts at 0x183, and is 11 bytes long. 0x182 may be
                   a continuation of the software type field, but I am most
                   likely wrong */
                string[11] = '\0';
                strncpy(string, &header[0x183], 11);
                printf("Product code and version: %s\n", string);
                printf("                Checksum: %02X%02X\n", header[0x18e],
                       header[0x18f]);
                string[16] = '\0';
                strncpy(string, &header[0x190], 16);
                printf("             I/O support: %s\n", string);
                /* the meaning of these fields may have been misinterpreted */
                printf("       ROM start address: %02X%02X%02X%02X\n",
                       header[0x1a0], header[0x1a1], header[0x1a2],
                       header[0x1a3]);
                printf("         ROM end address: %02X%02X%02X%02X\n",
                       header[0x1a4], header[0x1a5], header[0x1a6],
                       header[0x1a7]);
                /* is the modem data field really 20 bytes? XnaK's document
                   seems to indicate it is only 10... */
                string[20] = '\0';
                strncpy(string, &header[0x1bc], 20);                
                printf("              Modem data: %s\n", string);
                string[40] = '\0';
                strncpy(string, &header[0x1c8], 40);
                printf("                    Memo: %s\n", string);
                string[3] = '\0';
                strncpy(string, &header[0x1f0], 3);
                printf("               Countries: %s\n", string);

                fclose(fp);
        }
        else
                printf("show_info: Unrecognized file format\n");

        return 0;
}

/*****************************************************************************
 void smd_bin(unsigned char *, unsigned char *):
                Converts a 16KB SMD block to BIN.
 Input:         Destination buffer address, source buffer address.
 Returns:       Nothing.
*****************************************************************************/
void smd_bin(unsigned char *bin_block, unsigned char *smd_block)
{
        int i, o = 1, e = 0;

        /* convert 16KB of SMD to BIN */
        for (i = 0; i < 8192; i++)
        {
                bin_block[o] = smd_block[i];                
                bin_block[e] = smd_block[i+8192];
                o += 2;
                e += 2;
        }
}        
